/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include <cctype>
#include <stack>

#include "module.h"
#include "code.h"
#include "codeglob.h"
#include "perdroptions.h"
#include "rawdump.h"
#include "utils/compilecheck.hpp"

using namespace std;

// !!!
static bool commentBeginned = false;

// scrittura commenti
static void StartBeginComment()
{
  // scrivi inizio commento
  if (!commentBeginned)
    printf("\n");
  else
    printf("\n|\n");
  commentBeginned = true;

  printf("* ");
}

static void StartComment()
{
  if (!commentBeginned)
    StartBeginComment();
  else
    printf("\n|");
}

static void EndComment()
{
  if (commentBeginned)
  {
    printf("\n|\n");
    commentBeginned = false;
  }
}

// testa vari riferimenti
class CheckRef
{
public:
  CheckRef():count(0), countf(0), countb(0), flowref(FLOW_NONE) { }
  void operator()(const IstrReference& ref)
  {
    if (!ref.direct && !ref.isApi)
    {
      char type = '?';
      switch(ref.type)
      {
      case FLOW_CALL:  if (options.showcRef) type = 'c'; break;
      case FLOW_JUMP:  if (options.showURef) type = 'U'; break;
      case FLOW_CJUMP: if (options.showCRef) type = 'C'; break;
      }
      if (type != '?')
      {
        if (count==0)
        {
          StartBeginComment();
          if (!options.showURef && !options.showCRef)
            printf("Referenced by a call at Address:");
          else
            printf("Referenced by an (U)nconditional or (C)onditional Jump "
                 "or (c)all at Address:");
        }
        if ( (count%4) == 0 )
          StartComment();
        else
          printf(",");
        if (!options.showURef && !options.showCRef)
          printf(" %08X",ref.to);
        else
          printf(" %08X(%c)",ref.to,type);
        ++count;
        flowref = FLOW_RET;
      }
      else
      {
        if (flowref != FLOW_RET)
          flowref = ref.type;
        if (ref.to > ref.from) ++countf;
        if (ref.to < ref.from) ++countb;
      }
    }
    else if (ref.type == FLOW_CONST && flowref == FLOW_NONE)
      flowref = ref.type;
  }
  void Print()
  {
    if (flowref != FLOW_NONE && flowref != FLOW_RET)
    {
      if ((flowref == FLOW_CALL || flowref == FLOW_CONST) && !commentBeginned)
        printf("\n");
      EndComment();
      char buf[32], c;
      char* p = buf+1;
      if (flowref == FLOW_CALL)
      {
        strcpy(buf,"========");
        c = '=';
      }
      else
      {
        strcpy(buf,"--------");
        c = '-';
      }
      if (countb)
      {
        p += sprintf(p,"%d<",countb);
        *p++ = c;
      }
      if (countf)
      {
        p += sprintf(p,"%d>",countf);
        *p++ = c;
      }
      if (p >= buf+8)
        *p = '\0';
      printf("%s\n",buf);
    }
  }
private:
  int count;
  int countf, countb, flowref;
};

void CodeParser::CheckReference(vma_t address)
{
  // entry point
  if (address == module->GetEntryPoint())
  {
    StartBeginComment();
    printf("Program entry point");
  }

  // export
  // !!! manca ordinal
  pair<Symbols::iterator,Symbols::iterator> sym = module->GetExportSymbols()[address];
  for(;sym.first != sym.second;++sym.first)
  {
    StartBeginComment();
    printf("Exported function %s",(*sym.first).name.c_str());
  }

  // custom alias
  if (fnapi.aliases.IsValid(address))
  {
    pair<Symbols::iterator,Symbols::iterator> sym = fnapi.aliases[address];
    for(;sym.first != sym.second;++sym.first)
    {
      StartBeginComment();
      printf("Aliased function %s",(*sym.first).name.c_str());
    }
  }

  // check for references
  CheckRef refs = 
    std::for_each(istrReferences.beginReferenceTo(address),
                  istrReferences.endReferenceTo(address),
                  CheckRef());
  refs.Print();
}

// get call
class CallRef
{
public:
  CallRef():count(0),address(0) {};
  void operator()(const IstrReference& ref)
  {
    if (!ref.direct || !ref.isApi || ref.type != FLOW_CALL)
      return;
    if (ref.to != address)
    {
      ++count;
      address = ref.to;
      isApi   = ref.isApi;
    }
  }
  vma_t GetReference() const { return count == 1 ? address : 0; };
private:
  int count;
  vma_t address;
  bool isApi;
};

string CodeParser::GetCallReference(vma_t address, const Instruction& instruction)
{
  vector<string> refs;
  GetJumpReference(address, instruction, refs);
  if (!refs.size())
    return "";
  return refs[0];
}

void CodeParser::GetJumpReference(vma_t address, const Instruction& instruction, vector<string>& refs)
{
  // check for references
  CallRef ref = std::for_each(istrReferences.beginReferenceTo(address), istrReferences.endReferenceTo(address), CallRef());
  vma_t refAddress = ref.GetReference();
  if (refAddress)
  {
    pair<Symbols::iterator,Symbols::iterator> sym = module->GetSymbols()[refAddress];
    for (; sym.first != sym.second; ++ sym.first)
      refs.push_back((*sym.first).name);
  }

  // check for api reference
  for (int i=0; i<instruction.numArg; ++i)
  {
    const Param &arg = instruction.Args[i];
    if (arg.type == Param::t_memory)
    {
      // call/jmp dword ptr [constant]
      if (arg.mem_reg1 == null_reg && arg.mem_reg2 == null_reg)
      {
        pair<Symbols::iterator,Symbols::iterator> sym = module->GetSymbols()[arg.literal];
        for(;sym.first != sym.second;++sym.first)
          refs.push_back((*sym.first).name);
      }
    }
    if (arg.type == Param::t_literal)
    {
      // salti a jmp ad api
      if (IsPresent(apiAlias,arg.literal))
      {
        pair<Symbols::iterator,Symbols::iterator> sym = module->GetSymbols()[apiAlias[arg.literal]];
        for(;sym.first != sym.second;++sym.first)
          refs.push_back((*sym.first).name);
      }
      else if (fnapi.aliases.IsValid(arg.literal))
      {
        pair<Symbols::iterator,Symbols::iterator> sym = fnapi.aliases[arg.literal];
        for(;sym.first != sym.second;++sym.first)
          refs.push_back((*sym.first).name);
      }

      // jump to exported symbols (or any detected symbols in out code)
      pair<Symbols::iterator,Symbols::iterator> sym = module->GetExportSymbols()[arg.literal];
      for(;sym.first != sym.second;++sym.first)
        refs.push_back((*sym.first).name);
    }
  }
}

void CodeParser::CheckJumpReference(vma_t address, const Instruction& instruction)
{
  vector<string> refs;
  GetJumpReference(address, instruction, refs);

  for (vector<string>::iterator i = refs.begin(); i != refs.end(); ++i)
  {
    StartBeginComment();
    printf("Reference to %s", (*i).c_str());
  }
}

// !!! add code for other data, not only string
void CodeParser::CheckDataReference(vma_t address,const Instruction& instruction,std::string& comment)
{
  for (int i=0; i<instruction.numArg; ++i)
  {
    const Param &arg = instruction.Args[i];

    // if argument is a register ignore
    if (arg.type == Param::t_registry || arg.type == Param::t_farliteral)
      continue;
    
    vma_t address = arg.literal;
    if (arg.type == Param::t_memory)
    {
      if (!module->IsValid(address))
        continue;

      try
      {
        int type = byteInfo[address].GetType();
        if (type == ByteInfo::typeInteger || type == ByteInfo::typeFloat)
        {
          ObjectModule::DataReader reader = module->GetDataReader(address);
          char buf[32];
          unsigned len = byteInfo.GetLen(address);
          *buf = '\0';
          if (type == ByteInfo::typeInteger)
          {
            uint32_t i = 0;
            switch (len)
            {
            case 1: i = reader.ReadU8();    break;
            case 2: i = reader.ReadU16LE(); break;
            case 4: i = reader.ReadU32LE(); break;
            }
            if (i != 0)
            {
              int s = sprintf(buf,"%0*X (%d)",len*2,i,i);
              if (i < 0x7f && isprint(i))
                sprintf(buf+s," ('%c')",i);
              else
              {
                uint8_t c1, c2, c3, c4;
                c1 = (uint8_t)i;
                c2 = (uint8_t)(i>>8);
                c3 = (uint8_t)(i>>16);
                c4 = (uint8_t)(i>>24);
                if (isprint(c1) && isprint(c2))
                {
                  if (c3 == 0 && c4 == 0)
                    sprintf(buf+s," ('%c%c')",c1,c2);
                  else if (isprint(c3))
                  {
                    if (c4 == 0)
                      sprintf(buf+s," ('%c%c%c')",c1,c2,c3);
                    else if (isprint(c4))
                      sprintf(buf+s," ('%c%c%c%c')",c1,c2,c3,c4);
                  }
                }
              }
            }
          }
          else // type == ByteInfo::typeFloat
          {
            union {
              struct {
                uint32_t i4;
                uint32_t i8;
                uint16_t i10;
              };
              float f4;
              double f8;
              long double f10;
            } i;
            switch (len)
            {
            case 4:
              i.i4 = reader.ReadU32LE();
              if (i.f4 != 0.0)
                sprintf(buf,"%.7g",i.f4);
              break;
            case 8:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              if (i.f8 != 0.0)
                sprintf(buf,"%.7g",i.f8);
              break;
            case 10:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              i.i10 = reader.ReadU16LE();
              if (i.f10 != 0.0)
                sprintf(buf,"%.7Lg",i.f10);
              break;
            }
          }
          if (*buf)
          {
            if (!comment.empty())
              comment += " = ";       // argument = value
            comment += buf;
          }
          if (type != ByteInfo::typeInteger || len != 4)
            continue;
        }

        ObjectModule::DataReader reader = module->GetDataReader(address);
        address = reader.ReadU32LE();
      }
      catch(const OutOfAddress&)
      { continue; }
    }
    
    if (!module->IsValid(address))
      continue;

    unsigned len = byteInfo.GetLen(address);
    ByteInfo::TType type = byteInfo[address].GetType();
    if (type == ByteInfo::typeAnsiZ || type == ByteInfo::typeUnicodeZ ||
        IsPossibleString(address, len, type))
    {
      StartBeginComment();
      printf("Possible string reference %04Xh ",address);
      WriteString(address, len, type, 100);
    }
  }
}

struct FnCall
{
  const TFnArgs* args;
  int   numArg;
  bool  push;
  FnCall(const TFnArgs* a, int n, bool p) : args(a), numArg(n), push(p) {}

  void write(int arg, std::string& comment, int level)
  {
    while (level--)
      comment += "  ";
    if (arg == 0)
    {
      if (args->size() > 10)
        comment += ' ';
      comment += "r: ";
    }
    else
    {
      char buf[8];
      sprintf(buf,"%*d: ", (args->size() > 10)?2:1,arg);
      comment += buf;
    }
    comment += (*args)[arg];
  }
};

void CodeParser::CheckArgReference(vma_t address,const Instruction& instruction, std::string& comment)
{
  static stack<FnCall, vector<FnCall> > fn;
#define level (fn.size()-1)

  if (IsPresent(argFirst,address))
  {
    fn.push(FnCall(argFirst[address].first,
                   argFirst[address].second,
                   (instruction.instruction == istr_push)));
  }
  if (fn.size())
  {
    if (instruction.instruction == istr_call)
    {
      fn.top().write(0, comment, level);
      fn.pop();
    }
    else if (fn.top().push)
    {
      if (instruction.instruction == istr_push)
      {
        fn.top().write(fn.top().numArg, comment, level);
        --fn.top().numArg;
      }
    }
    else
    {
      const Param& arg = instruction.Args[0];
      if (instruction.instruction == istr_mov
          && arg.type == Param::t_memory
          && arg.mem_reg1 == REG(esp) && arg.mem_reg2 == null_reg)
      {
        int numArg = arg.literal/4 + 1;
        if (numArg >= 0 && numArg < fn.top().args->size())
        {
          fn.top().write(numArg, comment, level);
        }
      }
    }
  }
#undef level
}

// stampa una stringa
void CodeParser::WriteBinary(vma_t address,int len) const
{
  EndComment();

  _PRG_ASSERT(&*module != NULL);
  _PRG_ASSERT(len>0);
  ObjectModule::DataReader reader = module->GetDataReader(address);
  uint8_t buffer[256];
  while (len > 0)
  {
    unsigned left = (len > 256) ? 256u : (unsigned) len;
    for (unsigned n = 0; n < left; ++n)
      buffer[n] = reader.ReadU8();
    ::WriteBinary(address,buffer,left);
    address += left;
    len -= left;
  }
}

// scrive inizio istruzione/BYTE/DWORD come da inizio
// legge da ::module
void CodeParser::WriteBegin(vma_t address,int num_bytes,bool truncate) const
{
  int x=0,c;
  EndComment();
  printf("%08lX ",(long int)address);
  if (::options.showBytes)
  {
    if (truncate)
    {
      if (num_bytes > 10)
        num_bytes = 9;
      else
        truncate = false;
    }
    ObjectModule::DataReader reader = module->GetDataReader(address);
    for (int i=0;i<num_bytes;++i)
    {
      if ( x == 10 && num_bytes != 11)
      {
        x = 0;
        if (i != 0)
          printf("\n         ");
      }
      c = reader.ReadU8();
      printf("%02X",c);
      ++x;
    }

    if (truncate)
      printf("..");
    else for (;x<10;++x)
      printf("  ");
  }

  printf("\t");
}

// output a string constant
int CodeParser::WriteString(vma_t address, int len, ByteInfo::TType type, int truncate_at) const
{
  ObjectModule::DataReader reader = module->GetDataReader(address);
  bool uc = (type == ByteInfo::typeUnicodeZ);
  if (uc) {
    printf("L");
    len /= 2;
  }
  printf("\"");
  // let's not go overboard...
  const char* truncated = "";
  if (truncate_at > 0 && len > truncate_at)
  {
    len = truncate_at;
    truncated = "...";
  }
  while (--len > 0)
  {
    int c = reader.ReadU8();
    if (uc)
      c |= reader.ReadU8() << 8;
    switch(c)
    {
    case '\t':
      printf("\\t");
      break;
    case '\r':
      printf("\\r");
      break;
    case '\n':
      printf("\\n");
      break;
    case '\"':
      printf("\\\"");
      break;
    default:
      if (c < 256 && isprint(c))
        printf("%c",c);
      else if (c < 256)
        printf("\\x%02x",c);
      else
        printf("\\u%04x", c);
    }
  }
  printf("\"%s", truncated);

  return (uc) ? 2 : 1;
}

static void WritePriority(ByteInfo::TPriority priority)
{
  const char *s = "unknown";
  switch (priority)
  {
  case ByteInfo::priNone:
    s = "none"; break;
  case ByteInfo::priCheckOnly:
    s = "check only"; break;
  case ByteInfo::priFiller:
    s = "filler"; break;
  case ByteInfo::priConstant:
    s = "constant"; break;
  case ByteInfo::priApi:
    s = "api"; break;
  case ByteInfo::priExport:
    s = "export"; break;
  case ByteInfo::priHeuristics:
    s = "heuristics"; break;
  case ByteInfo::priSafeHeuristics:
    s = "safe heuristics"; break;
  case ByteInfo::priSafeExport:
    s = "safe export"; break;
  case ByteInfo::priEntryPoint:
    s = "entry point"; break;
  }
  StartBeginComment();
  printf("Current priority: %s",s);
}

// check conditions are in a range
#define COND_RANGE(s) \
  (istr_cmov##s >= istr_cmova && istr_cmov##s <= istr_cmovs \
   && istr_j##s >= istr_ja && istr_j##s <= istr_js \
   && istr_set##s >= istr_seta && istr_set##s <= istr_sets)
COMPILE_CHECK(conditionRanges,
  COND_RANGE(a) &&
  COND_RANGE(be) &&
  COND_RANGE(c) &&
  COND_RANGE(nc) &&
  COND_RANGE(z) &&
  COND_RANGE(nz) &&
  COND_RANGE(b) &&
  COND_RANGE(ae) &&
  COND_RANGE(e) &&
  COND_RANGE(ne) &&
  COND_RANGE(g) &&
  COND_RANGE(ge) &&
  COND_RANGE(l) &&
  COND_RANGE(le) &&
  COND_RANGE(no) &&
  COND_RANGE(ns) &&
  COND_RANGE(o) &&
  COND_RANGE(pe) &&
  COND_RANGE(po) &&
  COND_RANGE(s) &&
  COND_RANGE(nb) &&
  COND_RANGE(nbe) &&
  COND_RANGE(nl) &&
  COND_RANGE(nle) &&
  COND_RANGE(np)
);
#undef COND_RANGE

#define CFOLLOW(s1,s2) \
  ((istr_cmov##s1 + 1)   == istr_cmov##s2 \
   && (istr_j##s1 + 1)   == istr_j##s2 \
   && (istr_set##s1 + 1) == istr_set##s2)
#define IFOLLOW(s1,s2) \
  ((istr_##s1 + 1)   == istr_##s2)
// check instruction where we do arith
COMPILE_CHECK(istrSequences,
  // conditions
  CFOLLOW(c ,nc) &&
  CFOLLOW(nc,z ) &&
  CFOLLOW(z ,nz) &&
  CFOLLOW(nz,b ) &&
  CFOLLOW(b ,ae) &&
  CFOLLOW(ae,e ) &&
  CFOLLOW(e ,ne) &&
  // compares
  IFOLLOW(scasb, scasw) &&
  IFOLLOW(scasw, scasd) &&
  IFOLLOW(cmp,     cmpsb    ) &&
  IFOLLOW(cmpsb,   cmpsw    ) &&
  IFOLLOW(cmpsw,   cmpsd    ) &&
  IFOLLOW(cmpsd,   cmpxchg  ) &&
  IFOLLOW(cmpxchg, cmpxchg8b)
);
#undef IFOLLOW
#undef CFOLLOW

void CodeParser::WriteCode(vma_t writeStart, vma_t writeEnd)
{
  bool useCmp = false;
  vma_t end;

  // disassembly vero e proprio
  FOR_EACH_SECTION_BEGIN(module,p)
    bool isCode = (*p).IsCode();

    InstructionDecoder decoder(module->GetRelocationInfos());

    // init as invalid priority
    ByteInfo::TPriority priority = ByteInfo::TPriority(-1);

    vma_t pc = (*p).begin;
    end      = (*p).end;

    if (writeStart || writeEnd)
    {
      if (pc < writeStart)
        pc = writeStart;
      if (end > writeEnd)
        end = writeEnd;
    }

    for(; pc<end; )
    {
      ObjectModule::DataReader reader = module->GetDataReader(pc);
      int c;

      // testa cambio priorita
      if (options.showPriority)
      {
        ByteInfo::TPriority currPriority = byteInfo[pc].GetPriority();
        if (currPriority != priority)
        {
          WritePriority(currPriority);
          priority = currPriority;
        }
      }

      // test for reference
      CheckReference(pc);

      // testa se tipo conosciuto
      unsigned len = byteInfo.GetLen(pc);
      bool isInstruction = true;

      if (len != 0)
      {
        isInstruction = false;
        const ByteInfo& info = byteInfo[pc];
        switch (info.GetType())
        {
        case ByteInfo::typeInteger:
          {
            uint32_t i;
            switch (len)
            {
            case 1:
              i = reader.ReadU8();
              WriteBegin(pc,len);
              printf("BYTE\t%02X\n",i);
              break;
            case 2:
              i = reader.ReadU16LE();
              WriteBegin(pc,len);
              printf("WORD\t%04X\n",i);
              break;
            case 4:
              i = reader.ReadU32LE();
              WriteBegin(pc,len);
              printf("DWORD\t%08X\n",i);
              break;

            default:
              // !!! scrive come bytes normali
              //_PRG_ASSERT(0);
              WriteBinary(pc,len);
            }
            pc += len;
          }
          break;
        case ByteInfo::typeFloat:
          {
            // !!! not that portable
            union {
              struct {
                uint32_t i4;
                uint32_t i8;
                uint16_t i10;
              };
              float f4;
              double f8;
              long double f10;
            } i;
            switch (len)
            {
            case 4:
              i.i4 = reader.ReadU32LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("DWORD\t%#.7g\n",i.f4);
              break;
            case 8:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("QWORD\t%#.7g\n",i.f8);
              break;
            case 10:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              i.i10 = reader.ReadU16LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("TWORD\t%#.7Lg\n",i.f10);
              break;

            default:
              // !!! scrive come bytes normali
              //_PRG_ASSERT(0);
              WriteBinary(pc,len);
            }
            pc += len;
          }
          break;
        case ByteInfo::typePointer:
          {
            _PRG_ASSERT(len == (unsigned int)addr_bytes);
            vma_t address = reader.ReadU32LE();
            WriteBegin(pc,len);
            printf("DWORD\t%08lX\n",(long int)address);
            pc += len;
          }
          break;
        case ByteInfo::typeLoader:
          StartBeginComment();
          printf("Loader bytes");
          WriteBinary(pc,len);
          pc += len;
          break;
        case ByteInfo::typeAnsiZ:
        case ByteInfo::typeUnicodeZ:
          WriteBegin(pc, len, true);
          printf("StringZ\t");
          WriteString(pc, len, info.GetType());
          printf("\n");
          pc += len;
          break;
        default:
          isInstruction = true;
        }
      }

      if (isInstruction && !isCode)
      {
        len = 1;
        while (pc+len < end && byteInfo[pc+len].GetType() == ByteInfo::typeNone)
          ++len;
        WriteBinary(pc,len);
        pc += len;
        continue;
      }

      if (isInstruction)
      {
        Instruction currInstruction;

        // read next instruction
        int result = decoder.Decode(currInstruction,reader);

        if ( !result || (byteInfo.GetIstrLen(pc)==0 &&
             byteInfo.IsOccupied(pc,result)) )
        {
          reader = module->GetDataReader(pc);
          c = reader.ReadU8();
          WriteBegin(pc,1);
          printf("BYTE\t%02X\n",c);
          pc += 1;
        }
        else
        {
          int instr = currInstruction.instruction;
          if (useCmp)
          {
            if ((instr >= istr_ja    && instr <= istr_js)  ||
                (instr >= istr_seta  && instr <= istr_sets) ||
                (instr >= istr_cmova && instr <= istr_cmovs))
            {
              useCmp = false;
              if ((instr >= istr_jc    && instr <= istr_jnz)   ||
                  (instr >= istr_setc  && instr <= istr_setnz) ||
                  (instr >= istr_cmovc && instr <= istr_cmovnz))
              {
                currInstruction.instruction = instr + 4;
              }
            }
          }
          else if ((instr >= istr_cmp   && instr <= istr_cmpxchg8b) ||
                   (instr >= istr_scasb && instr <= istr_scasd))
            useCmp = true;

          std::string comment;
          CheckJumpReference(pc,currInstruction);
          CheckArgReference(pc, currInstruction, comment);
          CheckDataReference(pc, currInstruction, comment);
          WriteBegin(pc,result);
          WriteInstruction(currInstruction, comment);
          // go on
          pc += result;
        }
      }
    }
  FOR_EACH_SECTION_END(module,p)
}

void CodeParser::CheckForArgument(vma_t address,const Instruction& instr)
{
  const Param& arg = instr.Args[0];
  if (instr.instruction == istr_push ||
      (instr.instruction == istr_mov && arg.type == Param::t_memory
       && arg.mem_reg1 == REG(esp) && arg.mem_reg2 == null_reg))
  {
    argStack.push_back(ApiArgument(address,
                         (instr.instruction == istr_push) ? -1 : arg.literal));
  }
  else if (instr.instruction == istr_call)
  {
    const TFnArgs* args = 0;
    string ref = GetCallReference(address, instr);
    if (!ref.empty())
      args = fnapi.find(ref);
    else if (arg.type == Param::t_literal)
      args = fnapi.find(arg.literal);
    if (args)
    {
      if (args->size() == 1)
      {
        argFirst[address] = TArgsCnt(args,0);
      }
      else
      {
        int a, pos = argStack.size();
        for (a = args->size(); a > 1 && --pos >= 0; --a)
        {
          if (argStack[pos].offset != -1)
            if (argStack[pos].offset/4 + 1 >= args->size())
              --pos;
        }
        a = args->size() - (pos == -1 ? pos = 0, a : 1);
        if (a == 0)
          argFirst[address] = TArgsCnt(args,0);
        else
        {
          argFirst[argStack[pos].address] = TArgsCnt(args,a);
          argStack.erase(argStack.begin() + pos, argStack.end());
        }
      }
    }
    else
      argStack.clear();
  }
  else if (instr.instruction == istr_ret ||
           (instr.instruction == istr_jmp
            && (arg.type != Param::t_literal
                || arg.literal < address || arg.literal >= address+40)))
  {
    argStack.clear();
  }
}

void CodeParser::PostAnalysis()
{
  vma_t end;

  if (!fnapi.Populate())
    return;

  // disassembly vero e proprio
  FOR_EACH_SECTION_CODE_BEGIN(module,p)

    InstructionDecoder decoder(module->GetRelocationInfos());

    vma_t pc = (*p).begin;
    end      = (*p).end;

    unsigned len = 0;
    for(; pc<end; pc+=len)
    {
      ObjectModule::DataReader reader = module->GetDataReader(pc);
      Instruction currInstruction;

      // testa se tipo conosciuto
      len = byteInfo.GetLen(pc);

      if (len != 0)
      {
        if (byteInfo[pc].GetType() != ByteInfo::typeInstruction)
          continue;
      }

      // read next instruction
      len = decoder.Decode(currInstruction,reader);

      if ( !len || (byteInfo.GetIstrLen(pc)==0 &&
           byteInfo.IsOccupied(pc,len)) )
      {
        len = 1;
      }
      else
      {
        CheckForArgument(pc,currInstruction);
      }
    }
  FOR_EACH_SECTION_CODE_END(module,p)
}

// vim:tabstop=200:mouse=a:
